#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Annotate - Annotate functions given signature

"Annotate functions given signature"

# Extensively modified by CoolCat467

# Extensively modified from https://github.com/dropbox/pyannotate/blob/master/pyannotate_tools/annotations/parse.py

## Pyannotate is licensed under the terms of the Apache License, Version 2.0,
## reproduced below.
##   Copyright (c) 2017 Dropbox, Inc.
##
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

__title__ = 'Annotate'
__author__ = 'CoolCat467'
__license__ = 'Apache License, Version 2.0'


from collections.abc import Callable, Collection, Sequence
from tokenize import generate_tokens, tok_name
from typing import Any, Final, NoReturn
import re


BUILTINS: Final = {'List', 'Set', 'Type', 'Dict', 'Tuple'}


class ParseError(Exception):
    """Raised on any type comment parse error.
    The 'comment' attribute contains the comment that produced the error.
    """

    def __init__(self, comment: str | None = None) -> None:
        if comment is None:
            comment = ''
        elif hasattr(self, 'add_note'):  # New in Python 3.11
            self.add_note(comment)
        super().__init__(comment)


class Token:
    "Base token"
    __slots__ = ('text',)
    def __init__(self, text: str | None = None) -> None:
        self.text = text

    def __repr__(self) -> str:
        value = '' if self.text is None else f'{self.text!r}'
        return f'{self.__class__.__name__}({value})'


class Name(Token):
    """Name Token"""


class FunctionName(Token):
    """Function name"""


class ArgumentName(Name):
    """Argument name"""


class Operator(Token):
    """Operator Token"""


class DottedName(Name):
    """An identifier token, such as 'List', 'int' or 'package.name'"""


class ArgumentDefault(DottedName):
    """Argument value"""


class Separator(Operator):
    """A separator or punctuator token such as '(', '[' or '->'"""


class EndSeparator(Separator):
    """End separator that does not need extra space"""


class Colin(Separator):
    """Colin ':'"""


class TypeDef(Colin):
    """Type Definition Start ':'"""


class DefaultDef(Separator):
    """Argument Default Definition Start '='"""


class Keyword(Name):
    """Keyword such as 'async', 'def'"""


class ReturnTypeDef(Separator):
    """Return type definition '->'"""


class Definition(Keyword):
    """Definition keyword"""


class EndDefinition(Colin):
    """End Definition Colin"""


class End(Token):
    """A token representing the end of a type comment"""
    def __init__(self) -> None:
        super().__init__()


def tokenize(txt: str) -> list[Token]:
    """Translate a type comment into a list of tokens."""
    original = txt
    txt = txt.replace('?', '')
    tokens: list[Token] = []
    while True:
        if not txt:
            tokens.append(End())
            return tokens
        if txt[0] in {' ', '\n'}:
            txt = txt[1:]
        elif txt[0] in '()[],*':
            tokens.append(Separator(txt[0]))
            txt = txt[1:]
        elif txt[:2] == '->':
            tokens.append(Separator('->'))
            txt = txt[2:]
        elif txt[:3] == '...':
            tokens.append(DottedName('...'))
            txt = txt[3:]
        else:
            # lintcheck: invalid-name (C0103): Variable name "m" doesn't conform to snake_case naming style
            match_ = re.match(r'[-\w`]+(\s*(\.|:)\s*[-/\w]*)*', txt)
            if not match_:
                raise ParseError(f'Could not parse {txt!r} from {original!r}')
            fullname = match_.group(0)
            size = len(fullname)
            fullname = fullname.replace(' ', '')
            if '`' in fullname:
                fullname = fullname.split('`')[0]
            # pytz creates classes with the name of the timezone being used:
            # https://github.com/stub42/pytz/blob/f55399cddbef67c56db1b83e0939ecc1e276cf42/src/pytz/tzfile.py#L120-L123
            # This causes pyannotate to crash as it's invalid to have a class
            # name with a `/` in it (e.g. "pytz.tzfile.America/Los_Angeles")
            if fullname.startswith('pytz.tzfile.'):
                fullname = 'datetime.tzinfo'
            if '-' in fullname or '/' in fullname:
                print(f'{__file__}: {fullname} -> Any')
                # Not a valid Python name; there are many places that
                # generate these, so we just substitute Any rather
                # than crashing.
                fullname = 'Any'
            tokens.append(DottedName(fullname))
            txt = txt[size:]


def get_line_indent(text: str, char: str = ' ') -> int:
    "Return line indent"
    for idx, cur in enumerate(text.split(char)):
        if cur != '':
            return idx
    return 0


def tokenize_definition(line: int,
                        get_line: Callable[[int], str]) -> list[Token]:
    "Tokenize function definition"
    def read_line() -> str:
        "Incremental wrapper for get_line"
        nonlocal line
        value = get_line(line)
##        print(f'{value!r}')
        line += 1
        return value

    hasdef = False  # Do we have a definition token?
    defstart = False  # Has definition started?
    brackets = 0  # Current number of brackets in use
    typedef = 0  # Brackets in type definition (truthy if in type definition)
    default_arg = 0  # Brackets in default argument value
    indent = 0  # Current indent
    tokens: list[Token] = []

    # For each token tokenizer module finds
    for token in generate_tokens(read_line):
##        print(f'{tok_name[token.type]} {token.string!r}', end=' -> ')
        string = token.string
        if tok_name[token.type] == 'NAME':
            if string == 'async':  # Remember async
                tokens.append(Keyword(string))
            elif string == 'def':  # Error if more than one in case invalid start
                if hasdef:
                    raise ParseError('Did not expect second definition keyword')
                tokens.append(Definition(string))
                hasdef = True  # Have definition now
            elif not defstart:  # If not started definition, only function name
                tokens.append(FunctionName(string))
            elif typedef:  # If we are doing type definition, add DottedName
                # If last was also a DottedName and it ended with a dot,
                # instead add name text to previous
                if (tokens and isinstance(tokens[-1], DottedName)
                    and tokens[-1].text is not None
                    and tokens[-1].text.endswith('.')):
                    previous = tokens.pop().text
                    assert previous is not None
                    tokens.append(DottedName(previous + string))
                else:
                    tokens.append(DottedName(string))
            elif default_arg:
                # If defining argument default, add ArgumetDefault
                tokens.append(ArgumentDefault(string))
            else:  # Otherwise is an argument name
                tokens.append(ArgumentName(string))
        elif tok_name[token.type] == 'OP':
            if string in '([{':
                # We have traveled in an bracket level
                brackets += 1
                # If we have a definition and this is open parenthesis,
                # we are starting definition
                if not defstart and hasdef and string == '(':
                    defstart = True
                # If doing a type definition, we have typedef level
                if typedef:
                    typedef += 1
                if default_arg:
                    default_arg += 1
                tokens.append(EndSeparator(string))  # No spaces
            elif string in ')]}':
                # Left a bracket level
                brackets -= 1
                if typedef:  # Maybe left a type def level too
                    typedef -= 1
                if default_arg:
                    default_arg -= 1
                tokens.append(EndSeparator(string))
            elif string in {'*', '**'}:
                tokens.append(EndSeparator(string))
            elif string == '|':
                tokens.append(Separator(string))
            elif string == ',':
                # If exactly one level, leaving type definition land
                if typedef == 1:
                    typedef = 0
                if default_arg == 1:
                    default_arg = 0
                tokens.append(Separator(string))  # Space on end
            elif string == '->':
                # We are defining a return type
                tokens.append(ReturnTypeDef(string))
                typedef = 1
            elif string == ':':
                if defstart and not brackets:
                    # Ran out of brackets so done defining function
                    tokens.append(EndDefinition(string))
                    typedef = 0
                    break
                tokens.append(TypeDef(string))
                typedef = 1
            elif string == '=':  # Defining a default definition
                tokens.append(DefaultDef(string))
                typedef = 0  # shouldn't be defining type but make sure
                default_arg = 1
            elif string in {'-', '+', '~'}:  # Unary operators
                # Add on to previous ArgumentDefault if it exists
                if tokens and isinstance(tokens[-1], ArgumentDefault):
                    previous = tokens.pop().text
                    assert previous is not None
                    tokens.append(ArgumentDefault(previous + string))
                else:
                    tokens.append(ArgumentDefault(string))
            elif string in {'/', '//', '*', '^', '@', '%', '**',
                            '>>', '<<', '&', '|'}:  # All other math operators
                if tokens and isinstance(tokens[-1], ArgumentDefault):
                    prev = tokens.pop().text
                    assert prev is not None
                    tokens.append(ArgumentDefault(f'{prev} {string} '))
                elif string == '@' and not hasdef:
                    # Not matrix multiplication, decorator
                    tokens.append(Separator(string))
                else:
                    tokens.append(ArgumentDefault(string))
            elif string == '.' and (tokens
                                    and isinstance(tokens[-1], DottedName)):
                previous = tokens.pop().text
                assert previous is not None
                tokens.append(DottedName(previous + string))
            elif string == '...':  # Ellipsis constant
                tokens.append(DottedName(string))
            else:
                raise ParseError(f'Exaustive list of OP failed: {string!r}')
        elif tok_name[token.type] == 'INDENT':
            tokens.append(EndSeparator(string))
            indent = len(string)
        elif tok_name[token.type] in {'NL', 'NEWLINE'}:
            # replace separator ends with end separators
            if tokens and isinstance(tokens[-1], Separator):
                tokens.append(EndSeparator(tokens.pop().text))
            tokens.append(EndSeparator(string))
            # we have a new indent level
            indent = get_line_indent(get_line(line))
            tokens.append(EndSeparator(' '*indent))
        elif tok_name[token.type] == 'COMMENT':
            # Remember comments as separators
            tokens.append(EndSeparator(string))
        elif tok_name[token.type] in {'STRING', 'NUMBER'}:
            # Only argument default, so add on to previous if exists
            if tokens and isinstance(tokens[-1], ArgumentDefault):
                previous = tokens.pop().text
                assert previous is not None
                tokens.append(ArgumentDefault(previous + string))
            else:
                tokens.append(ArgumentDefault(string))
        else:
            raise ParseError(f'Unrecognized token type {tok_name[token.type]}')
##        print(tokens[-1])
##    print(tokens[-1])
##    print()
    tokens.append(End())
    return tokens


def get_type_repr(name: str) -> str:
    "Get representation of name"
    if '.' in name:
        module, text = name.split('.', 1)
        if module == 'typing' and text in BUILTINS:
            return text.lower()
    if ':' in name:
        module, text = name.split(':', 1)
        if module == 'typing' and text in BUILTINS:
            return text.lower()
##        return text
    return name


def get_typevalue_repr(typevalue: 'TypeValue') -> str:
    "Get representation of ClassType"
    name = get_type_repr(typevalue.name)
    if name in BUILTINS:
        name = name.lower()
    args = []
    for arg in typevalue.args:
        args.append(get_typevalue_repr(arg))
    if not args:
        return name
    if name == 'Union':
        return ' | '.join(args)
    values = ', '.join(args)
    return f'{name}[{values}]'


class TypeValue:
    """A type value, potentially collection of multiple"""
    __slots__ = ('name', 'args')
    def __init__(self,
                 name: str,
                 args: Sequence['TypeValue'] | None = None) -> None:
        self.name = name
        if args:
            self.args = tuple(args)
        else:
            self.args = ()

    def __repr__(self) -> str:
        return f'TypeValue({self.name!r}, {self.args!r})'

    def __str__(self) -> str:
        return get_typevalue_repr(self)


def list_or(values: Collection[str]) -> str:
    "Return comma separated listing of values joined with ` or `"
    if len(values) <= 2:
        return ' or '.join(values)
    copy = list(values)
    copy[-1] = f'or {copy[-1]}'
    return ', '.join(copy)


class Parser:
    """Implementation of the type comment parser"""
    __slots__ = ('tokens', 'i')
    def __init__(self, tokens: list[Token]) -> None:
        self.tokens = tokens
        self.i = 0

    def parse_type_list(self) -> list[TypeValue]:
        "Parse comma separated type list"
        types = []
        while self.lookup() not in (')', ']'):
            typ = self.parse_type()
            types.append(typ)
            string = self.lookup()
            if string == ',':
                self.expect(',')
            elif string not in {')', ']'}:
                self.fail(f'Expected ) or ], got {string!r}')
        return types

    def parse_union_list(self) -> list[TypeValue]:
        "Parse | separated union list"
        types = []
        while True:
            typ = self.parse_type()
            types.append(typ)
            if self.lookup() != '|':
                return types
            self.expect('|')

    def parse_single(self) -> TypeValue:
        "Parse single, does not look for | unions. Do not use directly."
        if self.lookup() == '[':
            # Callable[[...], ...]
            #          ^^^^^
            self.expect('[')
            args = self.parse_type_list()
            self.expect(']')
            return TypeValue('', args)
        token = self.expect_type(DottedName)
        if token.text == 'Any':
            return TypeValue('Any')
        if token.text == 'mypy_extensions.NoReturn':
            return TypeValue('NoReturn')
        if token.text == 'typing.NoReturn':
            return TypeValue('NoReturn')
        if token.text == 'Tuple':
            self.expect('[')
            args = self.parse_type_list()
            self.expect(']')
            return TypeValue(token.text, args)
        if token.text == 'Union':
            self.expect('[')
            items = self.parse_type_list()
            self.expect(']')
            if len(items) == 1:
                return items[0]
            if len(items) == 0:
                self.fail('No items in Union')
            return TypeValue(token.text, items)
        if self.lookup() == '[':
            self.expect('[')
            args = self.parse_type_list()
            self.expect(']')
            if token.text == 'Optional' and len(args) == 1:
                return TypeValue('Union', [args[0], TypeValue('None')])
            if token.text is None:
                token.text = 'None'
            return TypeValue(token.text, args)
        if token.text is None:
            token.text = 'None'
        return TypeValue(token.text)

    def parse_type(self) -> TypeValue:
        "Parse type including | unions"
        type_value = self.parse_single()
        if self.lookup() == '|':
            self.expect('|')
            args = [type_value] + self.parse_union_list()
            return TypeValue('Union', args)
        return type_value

    def fail(self, error: str | None) -> NoReturn:
        "Raise parse error"
        raise ParseError(error)

    def peek(self) -> Token:
        "Peek at next token"
        if self.i >= len(self.tokens):
            self.fail('Ran out of tokens')
        return self.tokens[self.i]

    def next(self) -> Token:
        "Get next token"
        token = self.peek()
        self.i += 1
        return token

    def expect(self, text: str) -> None:
        "Expect next token text to be text"
        if self.next().text != text:
            self.fail(f'Expected {text}')

    def expect_type(self,
                    token_type: type[Token] | tuple[type[Token], ...]
                    ) -> Token:
        "Expect next token to be instance of token_type. Return token."
        token = self.next()
        if not isinstance(token, token_type):
            if hasattr(token_type, '__iter__'):
                assert isinstance(token_type, tuple)
                expect_str = list_or([f'{cls.__name__!r}' for cls in token_type])
            else:
                expect_str = f'{token_type.__name__!r}'
            self.fail(f'Expected {expect_str}, got {token}')
        return token

    def lookup(self) -> str:
        "Peek at next token and return it's text"
        value = self.peek().text
        if value is None:
            return 'None'
        return value

    def back(self) -> None:
        "Go back one token"
        self.i -= 1

    def rest_tokens(self) -> list[Token]:
        "Return all tokens not processed"
        return self.tokens[self.i:]


def get_annotation(annotation: dict[str, Any],
                   get_line: Callable[[int], str]) -> str:
    "Get annotation"
##    print(f'{annotation = }\n')

    # Get definition tokens
    try:
        def_tokens = tokenize_definition(annotation['line'], get_line)
    except ParseError:
        print(f'Could not tokenize definition\n{annotation = }')
        raise
    except EOFError as exc:
        raise ParseError('Reached End of File, expected end of definition') from exc

    # Get the argument and return tokens from signature
    signature = annotation['signature']
    arg_tokens = [tokenize(arg) for arg in signature['arg_types']]
    return_tokens = tokenize(signature['return_type'])

    # Find start of argument definition
    new_lines = ''
    parser = Parser(def_tokens)
    while True:
        token = parser.next()
        if isinstance(token, ArgumentName):
            parser.back()
            break
        if token.text == ')':
            parser.back()
            break
        assert token.text is not None, "Unreachable, End token is only null text token"
        new_lines += token.text
        if isinstance(token, Keyword):
            new_lines += ' '

    # Find out which arguments we have to skip (*, **, /)
    skip_args: set[int] = set()
##    print(f'{parser.rest_tokens() = }\n')
    argument = 0
    argparser = Parser(parser.rest_tokens())
    while True:
        token = argparser.next()
        if isinstance(token, Separator):
            if token.text == ')':
                break
            if token.text == '/':
                skip_args.add(argument)
                argument += 1
            elif token.text == '*':
                skip_args.add(argument)
                argument += 1
                if argparser.lookup() == '*':
                    argparser.expect('*')
        elif isinstance(token, ArgumentName):
            argument += 1
        elif isinstance(token, End):
            parser.back()
            break

    arg_place = len(arg_tokens) - argument
    if arg_place < 0:
        # Self or class argument does not require type so annotations are
        # not given
        for skip in range(-arg_place):
            skip_args.add(skip)


    # Handle arguments
    argument = 0
    while True:
        token = parser.next()
        if isinstance(token, Separator):
            if isinstance(token, EndSeparator):
                assert token.text is not None
                new_lines += token.text
            else:
                new_lines += f'{token.text} '
            if token.text == ')':
                break
            if token.text == '*':
                argument += 1
        elif isinstance(token, ArgumentName):
            name = token.text
            type_text = ''
            if isinstance(parser.peek(), TypeDef):
                parser.expect(':')
                type_value = parser.parse_type()
                type_text = ': '+str(type_value)
            elif argument not in skip_args:
                type_value = Parser(arg_tokens[arg_place]).parse_type()
                type_text = ': '+str(type_value)
            if isinstance(parser.peek(), DefaultDef):
                parser.expect('=')
                type_text += ' = '
                if isinstance(parser.peek(), EndSeparator):
                    type_text += f'{parser.next().text}'
                    type_text += ', '.join(map(str, parser.parse_type_list()))
                    type_text += f'{parser.next().text}'
                else:
                    type_text += str(parser.parse_type())
            new_lines += f'{name}{type_text}'
            argument += 1
            arg_place += 1
        elif isinstance(token, End):
            print('Found End token during argument parsing')
            parser.back()
            break

    # Handle the end
    if isinstance(parser.peek(), EndDefinition):
        parser.tokens = [ReturnTypeDef('->')] + return_tokens[:-1] + parser.rest_tokens()
        parser.i = 0
    elif isinstance(parser.peek(), End):
        parser.tokens = [ReturnTypeDef('->')] + return_tokens[:-1] + parser.rest_tokens()
        parser.i = 0
    parser.expect('->')
    new_lines += ' -> '
    ret_type = str(parser.parse_type())
    new_lines += ret_type
    while True:
        token = parser.next()
        if isinstance(token, End):
            break
        assert token.text is not None
        new_lines += token.text

    return new_lines





def run() -> None:
    "Run test of module"
##    def get_line(line: int) -> str:
##        path = annotation['path']
##        assert isinstance(path, str), "Path must be string"
##        with open(path, 'r', encoding='utf-8') as files:
##            for line_no, line_text in enumerate(files):
##                if line_no == line-1:
##                    return line_text
##            raise EOFErrors
##    print(f'{get_annotation(annotation, get_line)!r}')


if __name__ == '__main__':
    print(f'{__title__}\nProgrammed by {__author__}.\n')
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Mypy Daemon Client - Modified version of mypy.dmypy.client

"""Client for mypy daemon mode.

This manages a daemon process which keeps useful state in memory
rather than having to read it back from disk on each run.

Modified version of mypy.dmypy.client
"""

# Modified by CoolCat467
# Original at https://github.com/python/mypy/blob/master/mypy/dmypy/client.py
# Original retrieved November 24th 2022

## Mypy (and mypyc) are licensed under the terms of the MIT license,
## reproduced below.
##
##    The MIT License
##
##    Copyright (c) 2012-2022 Jukka Lehtosalo and contributors
##    Copyright (c) 2015-2022 Dropbox, Inc.
##
##    Permission is hereby granted, free of charge, to any person obtaining a
##    copy of this software and associated documentation files (the "Software"),
##    to deal in the Software without restriction, including without limitation
##    the rights to use, copy, modify, merge, publish, distribute, sublicense,
##    and/or sell copies of the Software, and to permit persons to whom the
##    Software is furnished to do so, subject to the following conditions:
##
##    The above copyright notice and this permission notice shall be included in
##    all copies or substantial portions of the Software.
##
##    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
##    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
##    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
##    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
##    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
##    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
##    DEALINGS IN THE SOFTWARE.

from __future__ import annotations

__title__ = 'Mypy Daemon Client'
__license__ = 'MIT'

import json
import os
import time
from typing import Any, Sequence

from mypy.dmypy_os import alive as _alive, kill as _kill
from mypy.dmypy_util import receive as _receive
from mypy.ipc import IPCClient as _IPCClient, IPCException as _IPCException
from mypy.version import __version__
from mypy.dmypy_server import (
    daemonize as _daemonize,
    Server as _Server,
    process_start_options as _process_start_options
)

class BadStatus(Exception):
    """Exception raised when there is something wrong with the status file.

    For example:
    - No status file found
    - Status file malformed
    - Process whose process id is in the status file does not exist
    """

# Client-side infrastructure.

def _read_status(status_file: str) -> dict[str, object]:
    """Read status file.

    Raise BadStatus if the status file doesn't exist or contains
    invalid JSON or the JSON is not a dict.
    """
    if not os.path.isfile(status_file):
        raise BadStatus("No status file found")
    with open(status_file, 'r', encoding="utf-8") as file:
        try:
            data = json.load(file)
        except Exception as ex:
            raise BadStatus("Malformed status file (not JSON)") from ex
    if not isinstance(data, dict):
        raise BadStatus("Invalid status file (not a dict)")
    return data


def get_status(status_file: str) -> tuple[int, str]:
    """Read status file and check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatus if something's wrong.
    """
    data = _read_status(status_file)
    return _check_status(data)


def request(
    status_file: str,
    command: str, *, timeout: int | None = None, **kwds: object
) -> dict[str, Any]:
    """Send a request to the daemon.

    Return the JSON dict with the response.

    Raise BadStatus if there is something wrong with the status file
    or if the process whose process id is in the status file has died.

    Return {'error': <message>} if an IPC operation or receive()
    raised OSError.  This covers cases such as connection refused or
    closed prematurely as well as invalid JSON received.
    """
    response: dict[str, str] = {}
    args = dict(kwds)
    args["command"] = command
    # Tell the server whether this request was initiated from a human-facing terminal,
    # so that it can format the type checking output accordingly.
    args["is_tty"] = False
    args["terminal_width"] = 80
    bdata = json.dumps(args).encode("utf8")
    _, name = get_status(status_file)
    try:
        with _IPCClient(name, timeout) as client:
            client.write(bdata)
            response = _receive(client)
    except (OSError, _IPCException, ValueError, UnicodeError) as err:
        return {"error": str(err)}
    else:
        return response


def _check_status(data: dict[str, Any]) -> tuple[int, str]:
    """Check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatus if something's wrong.
    """
    if "pid" not in data:
        raise BadStatus("Invalid status file (no pid field)")
    pid = data["pid"]
    if not isinstance(pid, int):
        raise BadStatus("pid field is not an int")
    if not _alive(pid):
        raise BadStatus("Daemon has died")
    if "connection_name" not in data:
        raise BadStatus("Invalid status file (no connection_name field)")
    connection_name = data["connection_name"]
    if not isinstance(connection_name, str):
        raise BadStatus("connection_name field is not a string")
    return pid, connection_name


def is_running(status_file: str) -> bool:
    """Check if the server is running cleanly"""
    try:
        get_status(status_file)
    except BadStatus:
        return False
    return True


def stop(status_file: str) -> dict[str, Any]:
    """Stop daemon via a 'stop' request."""
    return request(status_file, "stop", timeout=5)


def _wait_for_server(status_file: str, timeout: float = 5.0) -> bool:
    """Wait until the server is up. Return False if timed out."""
    endtime = time.time() + timeout
    while time.time() < endtime:
        try:
            data = _read_status(status_file)
        except BadStatus:
            # If the file isn't there yet, retry later.
            time.sleep(0.1)
            continue
        # If the file's content is bogus or the process is dead, fail.
        try:
            _check_status(data)
        except BadStatus:
            return False
        return True
    return False


def _start_server(status_file: str,
                  *,
                  flags: list[str],
                  daemon_timeout: int | None = None,
                  allow_sources: bool = False,
                  log_file: str | None = None) -> bool:
    """Start the server and wait for it. Return False if error starting."""
    start_options = _process_start_options(flags, allow_sources)
    if _daemonize(start_options,
                  status_file,
                  timeout  = daemon_timeout,
                  log_file = log_file) != 0:
        return False
    return _wait_for_server(status_file)


def restart(status_file: str,
            *,
            flags: list[str],
            daemon_timeout: int | None = 0,
            allow_sources: bool = False,
            log_file: str | None = None) -> bool:
    """Restart daemon (it may or may not be running; but not hanging).

    Returns False if error starting."""
    try:
        stop(status_file)
    except BadStatus:
        # Bad or missing status file or dead process; good to start.
        pass
    return _start_server(status_file,
                         flags          = flags,
                         daemon_timeout = daemon_timeout,
                         allow_sources  = allow_sources,
                         log_file       = log_file)


# Action functions


def start(status_file: str,
          *,
          flags: list[str],
          daemon_timeout: int = 0,
          allow_sources: bool = False,
          log_file: str | None = None) -> bool:
    """Start daemon if not already running.

    Returns False if error starting / already running."""
    if not is_running(status_file):
        # Bad or missing status file or dead process; good to start.
        return _start_server(status_file,
                             flags          = flags,
                             daemon_timeout = daemon_timeout,
                             allow_sources  = allow_sources,
                             log_file       = log_file)
    return False


def status(status_file: str,
           *,
           timeout: int = 5,
           fswatcher_dump_file: str | None = None) -> dict[str, object]:
    """Ask daemon to return status."""
    return request(status_file, "status", timeout=timeout,
                   fswatcher_dump_file=fswatcher_dump_file)


def run(status_file: str,
        *,
        flags: list[str],
        timeout: int | None = None,
        daemon_timeout: int = 0,
        log_file: str | None = None,
        export_types: bool = False) -> dict[str, Any]:
    """Do a check, starting (or restarting) the daemon as necessary

    Restarts the daemon if the running daemon reports that it is
    required (due to a configuration change, for example)."""
    if not is_running(status_file):
        # Bad or missing status file or dead process; good to start.
        _start_server(status_file,
                      flags          = flags,
                      daemon_timeout = daemon_timeout,
                      allow_sources  = True,
                      log_file       = log_file)
    response = request(
        status_file,
        "run",
        timeout       = timeout,
        version       = __version__,
        args          = flags,
        export_types  = export_types,
    )
    # If the daemon signals that a restart is necessary, do it
    if "restart" in response:
        restart(status_file,
                flags          = flags,
                daemon_timeout = timeout,
                allow_sources  = True,
                log_file       = log_file)
        response = request(
            status_file,
            "run",
            timeout      = timeout,
            version      = __version__,
            args         = flags,
            export_types = export_types,
        )
    return response


def kill(status_file: str) -> bool:
    """Kill daemon process with SIGKILL. Return True if killed."""
    pid = get_status(status_file)[0]
    try:
        _kill(pid)
    except OSError as ex:
        print(ex)
        return False
    return True


def check(status_file: str,
          files: Sequence[str],
          *,
          timeout: int | None = None,
          export_types: bool = False) -> dict[str, Any]:
    """Ask the daemon to check a list of files."""
    return request(
        status_file,
        "check",
        timeout      = timeout,
        files        = files,
        export_types = export_types
    )


def recheck(status_file: str,
            export_types: bool,
            *,
            timeout: int | None = None,
            remove: list[str] | None = None,
            update: list[str] | None = None) -> dict[str, Any]:
    """Ask the daemon to recheck the previous list of files, with optional modifications.

    If at least one of --remove or --update is given, the server will
    update the list of files to check accordingly and assume that any other files
    are unchanged.  If none of these flags are given, the server will call stat()
    on each file last checked to determine its status.

    Files given in --update ought to exist.  Files given in --remove need not exist;
    if they don't they will be ignored.
    The lists may be empty but oughtn't contain duplicates or overlap.

    NOTE: The list of files is lost when the daemon is restarted.
    """
    if remove is not None or update is not None:
        return request(
            status_file,
            "recheck",
            timeout      = timeout,
            export_types = export_types,
            remove       = remove,
            update       = update,
        )
    return request(status_file,
                   "recheck",
                   timeout      = timeout,
                   export_types = export_types)


def inspect(status_file: str,
            location: str,  # line:col
            show: str = 'type',  # type, attrs, definition
            *,
            timeout: int | None = None,
            verbosity: int = 0,
            limit: int = 0,
            include_span: bool = False,
            include_kind: bool = False,
            include_object_attrs: bool = False,
            union_attrs: bool = False,
            force_reload: bool = False) -> dict[str, Any]:
    """Ask daemon to print the type of an expression."""
    return request(
        status_file,
        "inspect",
        timeout              = timeout,
        show                 = show,
        location             = location,
        verbosity            = verbosity,
        limit                = limit,
        include_span         = include_span,
        include_kind         = include_kind,
        include_object_attrs = include_object_attrs,
        union_attrs          = union_attrs,
        force_reload         = force_reload,
    )


def suggest(status_file: str,
            function: str,
            do_json: bool,
            *,
            timeout: int | None = None,
            callsites: bool = False,
            no_errors: bool = False,
            no_any: bool = False,
            flex_any: float | None = None,
            use_fixme: str | None = None,
            max_guesses: int | None = 64) -> dict[str, Any]:
    """Ask the daemon for a suggested signature."""
    return request(
        status_file,
        "suggest",
        timeout     = timeout,
        function    = function,
        json        = do_json,
        callsites   = callsites,
        no_errors   = no_errors,
        no_any      = no_any,
        flex_any    = flex_any,
        use_fixme   = use_fixme,
        max_guesses = max_guesses,
    )


def hang(status_file: str,
         *,
         timeout: int = 1) -> dict[str, Any]:
    """Hang for 100 seconds, as a debug hack."""
    if not isinstance(timeout, int):
        raise ValueError('Timeout must be an integer!')
    return request(status_file, "hang", timeout=timeout)


def do_daemon(status_file: str,
              flags: list[str],
              daemon_timeout: int | None = None) -> None:
    """Serve requests in the foreground."""

    options = _process_start_options(flags, allow_sources=False)
    _Server(options, status_file, timeout=daemon_timeout).serve()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Idle Type Check - Use mypy to type check open file, then add comments to file.

"Type Check IDLE Extension"

# Programmed by CoolCat467


__title__     = 'idlemypyextension'
__author__    = 'CoolCat467'
__license__   = 'GPLv3'
__version__   = '0.1.0'
__ver_major__ = 0
__ver_minor__ = 1
__ver_patch__ = 0

from typing import Any, TypeVar, cast, Final
from collections.abc import Callable

import os
import sys
import json
import math
import re
import traceback
from functools import wraps
from tkinter   import messagebox, Event, Tk

from idlelib         import search, searchengine
from idlelib.config  import idleConf
from idlelib.pyshell import PyShellEditorWindow

from idlemypyextension import annotate

DAEMON_TIMEOUT_MIN: Final = 5
ACTION_TIMEOUT_MIN: Final = 5

_HAS_MYPY = True
try:
    from idlemypyextension import client
except ImportError:
    print(f'{__file__}: Mypy not installed!')
    _HAS_MYPY = False

def get_required_config(values: dict[str, str], bind_defaults: dict[str, str]) -> str:
    "Get required configuration file data"
    config = ''
    # Get configuration defaults
    settings = '\n'.join(f'{key} = {default}' for key, default in values.items())
    if settings:
        config += f"\n[{__title__}]\n{settings}"
        if bind_defaults:
            config += '\n'
    # Get key bindings data
    settings = '\n'.join(f'{event} = {key}' for event, key in bind_defaults.items())
    if settings:
        config += f"\n[{__title__}_cfgBindings]\n{settings}"
    return config

def check_installed() -> bool:
    "Make sure extension installed."
    # Get list of system extensions
    extensions = list(idleConf.defaultCfg['extensions'])
    ex_defaults = idleConf.defaultCfg['extensions'].file

    # Import this extension (this file),
    module = __import__(__title__)

    # Get extension class
    if not hasattr(module, __title__):
        print(f'ERROR: Somehow, {__title__} was installed improperly, no {__title__} class '\
              'found in module. Please report this on github.', file=sys.stderr)
        sys.exit(1)

    cls = getattr(module, __title__)

    # Get extension class keybinding defaults
    required_config = get_required_config(
        getattr(cls, 'values', {}),
        getattr(cls, 'bind_defaults', {})
    )

    # If this extension not in there,
    if __title__ not in extensions:
        # Tell user how to add it to system list.
        print(f'{__title__} not in system registered extensions!')
        print(f'Please run the following command to add {__title__} to system extensions list.\n')
        # Make sure line-breaks will go properly in terminal
        add_data = required_config.replace('\n', '\\n')
        # Tell them command
        print(f"echo -e '{add_data}' | sudo tee -a {ex_defaults}")
        print()
    else:
        print(f'Configuration should be good! (v{__version__})')
        return True
    return False

def get_line_selection(line: int) -> tuple[str, str]:
    "Get selection strings for given line"
    return f'{line}.0', f'{line+1}.0'

# Stolen from idlelib.searchengine
def get_line_col(index: str) -> tuple[int, int]:
    "Return (line, col) tuple of integers from 'line.col' string."
    line, col = map(int, index.split('.', 1)) # Fails on invalid index
    return line, col

def get_line_indent(text: str, char: str = ' ') -> int:
    "Return line indent."
    for idx, cur in enumerate(text.split(char)):
        if cur != '':
            return idx
    return 0

def ensure_section_exists(section: str) -> bool:
    "Ensure section exists in user extensions configuration. Return True if created."
    if not section in idleConf.GetSectionList('user', 'extensions'):
        idleConf.userCfg['extensions'].AddSection(section)
        return True
    return False

F = TypeVar('F', bound=Callable[..., Any])

def undo_block(func: F) -> F:
    "Mark block of edits as a single undo block."
    @wraps(func)
    def undo_wrapper(self: 'idlemypyextension', *args: Any, **kwargs: Any) -> Any:
        "Wrap function in start and stop undo events."
        self.text.undo_block_start()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.text.undo_block_stop()
    return cast(F, undo_wrapper)

def ensure_values_exist_in_section(section: str, values: dict[str, str]) -> bool:
    "For each key in values, make sure key exists. If not, create and set to value. "\
    "Return True if created any defaults."
    need_save = False
    for key, default in values.items():
        value = idleConf.GetOption('extensions', section, key,
                                   warn_on_default=False)
        if value is None:
            idleConf.SetOption('extensions', section, key, default)
            need_save = True
    return need_save

def get_search_engine_params(engine: searchengine.SearchEngine) -> dict[str, str | bool]:
    "Get current search engine parameters"
    return {
        name: getattr(engine, f'{name}var').get()
        for name in ('pat', 're', 'case', 'word', 'wrap', 'back')
    }

def set_search_engine_params(engine: searchengine.SearchEngine,
                             data: dict[str, str | bool]) -> None:
    "Get current search engine parameters"
    for name in ('pat', 're', 'case', 'word', 'wrap', 'back'):
        if name in data:
            getattr(engine, f'{name}var').set(data[name])

def get_fake_editwin() -> PyShellEditorWindow:
    "Get fake edit window for testing"
    from idlelib.pyshell import PyShellEditorWindow
    class FakeEditWindow(PyShellEditorWindow):
        "FakeEditWindow for testing"
        from tkinter import Text
        class _FakeText(Text):
            bind = lambda x, y: None
        text = _FakeText
        from idlelib.format import FormatRegion
        fregion = FormatRegion
        from idlelib.pyshell import PyShellFileList
        flist = PyShellFileList
        from idlelib.iomenu import IOBinding
        io = IOBinding
    return FakeEditWindow

# Important weird: If event handler function returns 'break',
# then it prevents other bindings of same event type from running.
# If returns None, normal and others are also run.

class idlemypyextension:# pylint: disable=invalid-name
    "Add comments from mypy to an open program."
    __slots__ = (
        'editwin',
        'text',
        'formatter',
        'files',
        'flist',
    )
    # Extend the file and format menus.
    menudefs = [
        ('edit', [
            None,
            ('_Type Check File', '<<type-check>>'),
            ('Find Next Type Comment', '<<find-next-type-comment>>')
        ] ),
        ('format', [
            ('Suggest Signature', '<<suggest-signature>>'),
            ('Remove Type Comments', '<<remove-type-comments>>')
        ] ),
        ('run', [
            ('Shutdown dmypy daemon', '<<shutdown-dmypy-daemon>>')
        ] )
    ]
    # Default values for configuration file
    values = {
        'enable'         : 'True',
        'enable_editor'  : 'True',
        'enable_shell'   : 'False',
        'daemon_flags'   : 'None',
        'search_wrap'    : 'True',
        'suggest_replace': 'False',
        'timeout_mins'   : '30',
        'action_max_sec' : '40',
    }
    # Default key binds for configuration file
    bind_defaults = {
        'type-check'            : '<Alt-Key-t>',
        'suggest-signature'     : '<Alt-Key-s>',
        'remove-type-comments'  : '<Alt-Shift-Key-T>',
        'find-next-type-comment': '<Alt-Key-g>'
    }
    comment = '# types: '

    # Overwritten in reload
    daemon_flags        = 'None'
    search_wrap         = 'True'
    suggest_replace     = 'False'
    timeout_mins = '30'
    action_max_sec      = '40'

    # Class attributes
    idlerc_folder = os.path.expanduser(idleConf.userdir)
    mypy_folder   = os.path.join(idlerc_folder, 'mypy')
    status_file   = os.path.join(mypy_folder, 'dmypy.json')
    log_file      = os.path.join(mypy_folder, 'log.txt')

    def __init__(self, editwin: PyShellEditorWindow) -> None:
        "Initialize the settings for this extension."
##        self.editwin  : idlelib.pyshell.PyShellEditorWindow = editwin
##        self.text     : idlelib.multicall.MultiCallCreator  = editwin.text
##        self.formatter: idlelib.format.FormatRegion         = editwin.fregion
##        self.flist    : idlelib.pyshell.PyShellFileList     = editwin.flist
##        self.files    : idlelib.iomenu.IOBinding            = editwin.io
        self.editwin   = editwin
        self.text      = editwin.text
        self.formatter = editwin.fregion
        self.flist     = editwin.flist
        self.files     = editwin.io

        if not os.path.exists(self.mypy_folder):
            os.mkdir(self.mypy_folder)

        for attr_name in (a for a in dir(self) if not a.startswith('_')):
            if attr_name.endswith('_event'):
                bind_name = '-'.join(attr_name.split('_')[:-1]).lower()
                self.text.bind(f'<<{bind_name}>>', getattr(self, attr_name))
##                print(f'{attr_name} -> {bind_name}')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.editwin!r})'

    @property
    def daemon_timeout(self) -> int:
        "Daemon timeout"
        if self.timeout_mins == 'None':
            return DAEMON_TIMEOUT_MIN * 60
        try:
            return max(DAEMON_TIMEOUT_MIN * 60,
                       math.ceil(float(self.timeout_mins) * 60))
        except ValueError:
            return DAEMON_TIMEOUT_MIN * 60

    @property
    def action_timeout(self) -> int | None:
        "Action timeout"
        if self.action_max_sec == 'None':
            return None
        try:
            return max(ACTION_TIMEOUT_MIN, int(self.action_max_sec))
        except ValueError:
            return max(ACTION_TIMEOUT_MIN, int(self.values['action_max_sec']))

    @property
    def flags(self) -> list[str]:
        "Daemon server flags"
        base = {
            f'--cache-dir={self.mypy_folder}',
##            '--cache-fine-grained',
            '--hide-error-context',
            '--no-color-output',
            '--show-absolute-path',
            '--no-error-summary',
        }
        if self.daemon_flags == 'None':
            return list(base)
        return list(base | {s.strip() for s in self.daemon_flags.split(' ')})

    @classmethod
    def ensure_bindings_exist(cls) -> bool:
        "Ensure key bindings exist in user extensions configuration. Return True if need to save."
        need_save = False
        section = f'{cls.__name__}_cfgBindings'
        if ensure_section_exists(section):
            need_save = True
        if ensure_values_exist_in_section(section, cls.bind_defaults):
            need_save = True
        return need_save

    @classmethod
    def ensure_config_exists(cls) -> bool:
        "Ensure required configuration exists for this extension. Return True if need to save."
        need_save = False
        if ensure_section_exists(cls.__name__):
            need_save = True
        if ensure_values_exist_in_section(cls.__name__, cls.values):
            need_save = True
        return need_save

    @classmethod
    def reload(cls) -> None:
        "Load class variables from configuration."
##        # Ensure file default values exist so they appear in settings menu
##        save = cls.ensure_config_exists()
##        if cls.ensure_bindings_exist() or save:
##            idleConf.SaveUserCfgFiles()

        # Reload configuration file
        idleConf.LoadCfgFiles()

        # For all possible configuration values
        for key, default in cls.values.items():
            # Set attribute of key name to key value from configuration file
            if not key in {'enable', 'enable_editor', 'enable_shell'}:
                value = idleConf.GetOption(
                    'extensions',
                    cls.__name__,
                    key,
                    default=default
                )
                setattr(cls, key, value)

    @classmethod
    def get_msg_line(cls, indent: int, msg: str) -> str:
        "Return message line given indent and message."
        strindent = ' '*indent
        return f'{strindent}{cls.comment}{msg}'

    def get_line(self, line: int) -> str:
        "Get the characters from the given line in the currently open file."
        chars: str = self.text.get(*get_line_selection(line))
        return chars

    def comment_exists(self, line: int, text: str) -> bool:
        "Return True if comment for message already exists on line."
        return self.get_msg_line(0, text) in self.get_line(line-1)

    def add_comment(self,
                    message: dict[str, str | int],
                    max_exist_up: int = 0) -> bool:
        "Return True if added new comment, False if already exists."
        # Get line and message from output
##        file = str(message['file'])
        line = int(message['line'])
        msg  = str(message['message'])

        # If there is already a comment from us there, ignore that line.
        # +1-1 is so at least up by 1 is checked, range(0) = []
        for i in range(max_exist_up+1):
            if self.comment_exists(line-(i-1), msg):
                return False

        # Get line checker is talking about
        chars = self.get_line(line)

        # Figure out line indent
        indent = get_line_indent(chars)

        # Add comment line
        chars = self.get_msg_line(indent, msg)+'\n'+chars

        # Save changes
        start, end = get_line_selection(line)
        self.text.delete(start, end)
        self.text.insert(start, chars, None)
        return True

    @staticmethod
    def parse_comments(comments: str,
                       default_file: str,
                       default_line: int
                       ) -> dict[str, list[dict[str, str | int]]]:
        "Get list of message dictionaries from mypy output."
        error_type = re.compile(r'  \[[a-z\-]+\]\s*$')

        files: dict[str, list[dict[str, str | int]]] = {}
        for comment in comments.splitlines():
            filename = default_file
            line     = default_line
            col      = 0
            mtype    = 'unrecognized'

            if comment.count(': ') < 2:
                text = comment
            else:
                where, mtype, text = comment.split(': ', 2)

                position = where.split(':')

                filename = position[0]
                if len(position) > 1:
                    line = int(position[1])
                if len(position) > 2:
                    col = int(position[2])
            comment_type = error_type.search(text)
            if comment_type is not None:
                text = text[:comment_type.start()]
                mtype = f'{comment_type.group(0)[3:-1]} {mtype}'

            message: dict[str, str | int] = {
                'file'   : filename,
                'line'   : line,
                'column' : col,
                'type'   : mtype,
                'message': f'{mtype}: {text}'
            }

            if not filename in files:
                files[filename] = []
            files[filename].append(message)
        return files

    def get_pointers(self,
                     messages: list[dict[str, int | str]]
                     ) -> dict[str, int | str] | None:
        "Return message pointing to message column position"
        line = int(messages[0]['line'])+1

        # Figure out line intent
        indent = get_line_indent(self.get_line(line))

        columns: list[int] = []
        for message in messages:
            columns.append(int(message['column']))

        lastcol = len(self.comment) + indent + 1
        new_line = ''
        for col in sorted(frozenset(columns)):
            if col < lastcol:
                continue
            spaces = col - lastcol
            new_line += ' '*spaces + '^'
            lastcol = col+1

        if not new_line.strip():
            return None

        return {
            'line'   : line,
            'message': new_line
        }

    def add_comments(self,
                     target_filename: str,
                     normal: str) -> list[int]:
        "Add comments for target filename, return list of comments added"
        start_line: int = self.editwin.getlineno()

        files = self.parse_comments(
            normal,
            os.path.abspath(self.files.filename),
            start_line
        )

        # Only handling messages for target filename
        line_data: dict[int, list[dict[str, Any]]] = {}
        if target_filename in files:
            for message in files[target_filename]:
                line = message['line']
                assert isinstance(line, int), "Line must be int"
                if not line in line_data:
                    line_data[line] = []
                line_data[line].append(message)

        line_order: list[int] = list(sorted(line_data, reverse=True))
        first: int = line_order[-1] if line_order else start_line

        if not first in line_data:# if used starting line
            line_data[first] = []
            line_order.append(first)

        for filename in {f for f in files if f != target_filename}:
            line_data[first].append({
                'file'   : target_filename,
                'line'   : first,
                'column' : 0,
                'type'   : 'note',
                'message': f'Another file has errors: {filename}'
            })

        comments = []
        for line in line_order:
            messages = line_data[line]
            if not messages:
                continue
            pointers = self.get_pointers(messages)
            if pointers is not None:
                messages.append(pointers)

            total = len(messages)
            for message in reversed(messages):
                if self.add_comment(message, total):
                    comments.append(line)
        return comments

    def ask_save_dialog(self) -> bool:
        "Ask to save dialog stolen from idlelib.runscript.ScriptBinding"
        msg = 'Source Must Be Saved\n' + 5*' ' + 'OK to Save?'
        confirm: bool = messagebox.askokcancel(
            title   = 'Save Before Run or Check',
            message = msg,
            default = messagebox.OK,
            parent  = self.text
        )
        return confirm

    def ensure_daemon_running(self) -> bool:
        "Make sure daemon is running. Return False if cannot continue"
        if not client.is_running(self.status_file):
            started = client.start(self.status_file,
                                   flags = self.flags,
                                   daemon_timeout = self.daemon_timeout,
                                   log_file = self.log_file)
            return started
        return True

    def shutdown_dmypy_daemon_event(self, event: 'Event[Any]') -> str:
        "Shutdown the dmypy daemon"
        # pylint: disable=unused-argument
        if not client.is_running(self.status_file):
            self.text.bell()
            return 'break'
        # Only stop if running
        response = client.stop(self.status_file)
        if any((v in response and response[v] for v in ('err', 'error'))):
            # Kill
            client.kill(self.status_file)
        return 'break'

    def check(self, file: str) -> dict[str, Any]:
        "Preform dmypy check"
        if not self.ensure_daemon_running():
            return {'out': '',
                    'err': 'Error: Could not start mypy daemon'}
        flags = self.flags
        flags += [file]
        return client.run(self.status_file,
                          flags          = flags,
                          timeout        = self.action_timeout,
                          daemon_timeout = self.daemon_timeout,
                          log_file       = self.log_file,
                          export_types   = True)

    def get_suggestion_text(self,
                            annotation: dict[str, Any]) -> str | None:
        """Get suggestion text from annotation.

        Return None on error or no difference, text if different"""
        while annotation['line'] >= 0 and not 'def' in self.get_line(annotation['line']):
            annotation['line'] -= 1
        line = annotation['line']

        try:
            text = annotate.get_annotation(annotation, self.get_line)
        except annotate.ParseError as ex:
            ex_text, ex_traceback = sys.exc_info()[1:]
            traceback.print_exception(
                None,  # Ignored since python 3.5
                value = ex_text,
                tb = ex_traceback,
                limit = None,
                chain = True
            )
            indent = get_line_indent(self.get_line(line))
            return self.get_msg_line(indent, f'Error generating suggestion: {ex}')

        select_start = f'{line}.0'
        line_end = line+len(text.splitlines())
        select_end = f'{line_end}.0'

        if text == self.text.get(select_start, select_end)[:-1]:
            return None
        return text

    def suggest(self, file: str, line: int) -> None:
        "Preform dmypy suggest"
        if not self.ensure_daemon_running():
            response = {'err': 'Error: Could not start mypy daemon'}
        else:
            function = f'{file}:{line}'
            response = client.suggest(self.status_file,
                                      function = function,
                                      do_json  = True,
                                      timeout  = self.action_timeout)
##        print(f'{response = }')

        errors = ''
        if 'error' in response:
            errors += response['error']
        if 'err' in response:
            errors += response['err']

        # Display errors
        if errors:
            lines = errors.splitlines()
            lines[0] = f'Error running mypy: {lines[0]}'
            for message in reversed(lines):
                self.add_comment({
                    'file': file,
                    'line': line,
                    'message': message,
                }, len(lines))

            self.text.bell()
            return

        annotations = json.loads(response['out'])

        line = annotations[0]['line']

        samples: dict[int, list[str]] = {}
        for annotation in annotations:
            count = annotation['samples']
            text = self.get_suggestion_text(annotation)
            if text is None:
                continue
            if not count in samples:
                samples[count] = []
            samples[count].append(text)

        order = sorted(samples, reverse=True)
        lines = []
        for count in order:
            for sample in samples[count]:
                if not sample in lines:
                    lines.append(sample)

        replace = self.suggest_replace == 'True'

        if len(lines) == 1:
            text = lines[0]
            if 'Error generating suggestion: ' in text:
                replace = False
        else:
            text = '\n'.join(lines)
            replace = False

        select_start = f'{line}.0'
        line_end = line+len(text.splitlines())
        select_end = f'{line_end}.0'

        if not text or text == self.text.get(select_start, select_end)[:-1]:
            # Bell to let user know happened, just nothing to do
            self.editwin.gotoline(line)
            self.text.bell()
            return

        if replace:
            self.text.delete(select_start, select_end)
        elif 'Error generating suggestion: ' not in text:
            text = '\n'.join(f'##{l}' for l in text.splitlines())
        text += '\n'

        self.text.insert(select_start, text, None)
        self.editwin.gotoline(line)
        self.text.bell()

    @undo_block
    def suggest_signature_event(self, event: 'Event[Any]') -> str:
        "Handle suggest signature event"
        # pylint: disable=unused-argument
        init_return, file, start_line_no = self.initial()
        if init_return is not None:
            return init_return

        self.suggest(file, start_line_no)

        return 'break'

    def initial(self) -> tuple[str | None, str, int]:
        """Do common initial setup. Return error or none, file, and start line

        Reload configuration, make sure file is saved,
        and make sure mypy is installed"""
        self.reload()

        # Get file we are checking
        file: str = os.path.abspath(self.files.filename)

        # Remember where we started
        start_line_no: int = self.editwin.getlineno()

        if not _HAS_MYPY:
            self.add_comment({
                'file': file,
                'line': start_line_no,
                'message': 'Could not import mypy. '\
                'Please install mypy and restart IDLE to use this extension.'
            }, start_line_no)

            # Make bell sound so user knows they need to pay attention
            self.text.bell()
            return 'break', file, start_line_no

        # Make sure file is saved.
        if not self.files.get_saved():
            if not self.ask_save_dialog():
                # If not ok to save, do not run. Would break file.
                self.text.bell()
                return 'break', file, start_line_no
            # Otherwise, we are clear to save
            self.files.save(None)
            self.files.set_saved(True)

        # Everything worked
        return None, file, start_line_no

    @undo_block
    def type_check_event(self, event: 'Event[Any]') -> str:
        "Preform a mypy check and add comments."
        # pylint: disable=unused-argument
        init_return, file, start_line_no = self.initial()
        if init_return is not None:
            return init_return

        # Run mypy on open file
        response = self.check(file)
        normal = ''
        if 'out' in response:
            normal = response['out']
        errors = ''
        if 'err' in response:
            errors = response['err']
        if 'error' in response:
            errors += response['error']
##        print(response)

        if errors:
            lines = errors.splitlines()
            lines[0] = f'Error running mypy: {lines[0]}'
            for message in reversed(lines):
                self.add_comment({
                    'file': file,
                    'line': start_line_no,
                    'message': message,
                }, len(lines))

            self.text.bell()
            return 'break'

        if normal:
            # Add code comments
            self.add_comments(file, normal)

        # Make bell sound so user knows we are done,
        # as it freezes a bit while mypy looks at the file
        self.text.bell()
        return 'break'

    def remove_type_comments_event(self, event: 'Event[Any]') -> str:
        "Remove selected mypy comments."
        # pylint: disable=unused-argument
        # Get selected region lines
        head, tail, chars, lines = self.formatter.get_region()
        if not self.comment in chars:
            # Make bell sound so user knows this ran even though
            # nothing happened.
            self.text.bell()
            return 'break'
        # Using dict so we can reverse and enumerate
        ldict = dict(enumerate(lines))
        for idx in sorted(ldict.keys(), reverse=True):
            line = ldict[idx]
            # If after indent there is mypy comment
            if line.lstrip().startswith(self.comment):
                # If so, remove line
                del lines[idx]
        # Apply changes
        self.formatter.set_region(head, tail, chars, lines)
        return 'break'

    @undo_block
    def remove_all_type_comments(self, event: 'Event[Any]') -> str:
        "Remove all mypy comments."
        # pylint: disable=unused-argument
        eof_idx = self.text.index('end')

        chars = self.text.get('0.0', eof_idx)

        lines = chars.splitlines()
        modified = False
        for idx in reversed(range(len(lines))):
            if lines[idx].lstrip().startswith(self.comment):
                del lines[idx]
                modified = True
        if not modified:
            return 'break'

        chars = '\n'.join(lines)

        # Apply changes
        self.text.delete('0.0', eof_idx)
        self.text.insert('0.0', chars, None)
        return 'break'

    @undo_block
    def find_next_type_comment_event(self, event: 'Event[Any]') -> str:
        "Find next comment by hacking the search dialog engine."
        # pylint: disable=unused-argument
        self.reload()

        root: Tk
        root = self.text._root()# pylint: disable=protected-access

        # Get search engine singleton from root
        engine: searchengine.SearchEngine = searchengine.get(root)

        # Get current search prams
        global_search_params = get_search_engine_params(engine)

        # Set search pattern to comment starter
        set_search_engine_params(engine, {
            'pat' : f'^\\s*{self.comment}',
            're'  : True,
            'case': True,
            'word': False,
            'wrap': self.search_wrap == 'True',
            'back': False
        })

        # Find current pattern
        search.find_again(self.text)

        # Re-apply previous search prams
        set_search_engine_params(engine, global_search_params)
        return 'break'

idlemypyextension.reload()

if __name__ == '__main__':
    print(f'{__title__} v{__version__}\nProgrammed by {__author__}.\n')
    check_installed()
##    self = idlemypyextension(get_fake_editwin())
